import pytest

from pkglts.option.doc.badge import Badge


def test_fmt_badge_works_correctly():
    badge = Badge(name="mybadge",
                  url="myurl",
                  url_img="myurl/img.svg",
                  text="mytxt",
                  )

    sec = badge.format('md')
    assert badge.url in sec
    assert badge.url_img in sec
    assert badge.text in sec

    sec = badge.format('rst')
    assert badge.url in sec
    assert badge.url_img in sec
    assert badge.text in sec

    with pytest.raises(UserWarning):
        badge.format('docx')



import logging
from pathlib import Path

from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionData(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        LOGGER.info("update parameters %s", self._name)
        sec = dict(
            filetype=[".json", ".ini"],
            use_ext_dir=False
        )
        cfg[self._name] = sec

    def require_option(self, cfg):
        return ['src']



"""
Base class for options.
"""
from pathlib import Path


class Option(object):
    """Base class to store information associated with an option
    """

    def __init__(self, name):
        self._name = name

    def version(self):
        """Current version of this option.

        Returns:
            (str): X.X.X
        """
        return "0.0.0"

    def root_dir(self):
        """Base directory containing option definition files.

        Returns:
            (Path)
        """
        return Path(__file__).parent

    def example_dir(self):
        """Directory containing option example files.

        Returns:
            (Path)
        """
        pth = self.root_dir() / 'example'
        if pth.exists():
            return pth

        return None

    def resource_dir(self):
        """Directory containing option resource files.

        Returns:
            (Path)
        """
        pth = self.root_dir() / 'resource'
        if pth.exists():
            return pth

        return None

    def update_parameters(self, cfg):
        """Update configuration with option parameters.

        Args:
            cfg (Config):  current package configuration

        Returns:
            (Config)
        """
        cfg[self._name] = {}

        return cfg

    def check(self, cfg):
        """Check validity of parameters for this option.

        Args:
            cfg (Config):  current package configuration

        Returns:
            (list of str): list of failing params
        """
        return []

    def require_option(self, cfg):
        """Names of other options required by this option.

        Args:
            cfg (Config):  current package configuration

        Returns:
            (list of str)
        """
        return []

    def require(self, cfg):
        """Check dependencies for this option.

        Args:
            cfg (Config):  current package configuration

        Returns:
            (list of Dependency): list of packages this option require
        """
        return []

    def environment_extensions(self, cfg):
        """Get jinja2 environment extensions defined by this option.

        Args:
            cfg (Config):  current package configuration

        Returns:
            (dict of str: any): extensions defined by this option
        """
        return {}

    def regenerate(self, *args, **kwds):
        """Call regenerate associated with this option.

        Returns:
            (any)
        """
        return None

    def tools(self, cfg):
        """Iterate on tools defined by this option.

        Args:
            cfg (Config):  current package configuration

        Returns:
            (iter of func): cli parser
        """
        return []



import logging
from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option.doc.badge import Badge
from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionCoveralls(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['coverage', 'travis']

    def require(self, cfg):
        del cfg
        yield Dependency('coveralls', intent='test')

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        url = f"coveralls.io/github/{owner}/{project}?branch=master"
        img = f"coveralls.io/repos/github/{owner}/{project}/badge.svg?branch=master"
        badge = Badge(
            name="coveralls",
            url=url,
            url_img=img,
            text="Coverage report status"
        )

        return {"badge": badge}



import logging
from pathlib import Path

from pkglts.option_object import Option
from pkglts.version import __version__
from . import nbcompile

LOGGER = logging.getLogger(__name__)


class OptionNotebook(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        LOGGER.info("update parameters %s", self._name)
        sec = dict(
            src_directory="example"
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        src_directory = Path(cfg[self._name]['src_directory'])

        if not src_directory.exists():
            invalids.append("notebook.src_directory")

        return invalids

    def require_option(self, cfg):
        return ['sphinx']

    def tools(self, cfg):
        del cfg
        yield nbcompile.parser_nbcompile



import pytest
from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.manage_tools import check_option_parameters


def test_check_handle_option_config_not_existing():
    cfg = Config(DEFAULT_CFG)
    with pytest.raises(KeyError):
        check_option_parameters("toto", cfg)



import pytest
from pkglts.config_management import Config
from pkglts.option.coveralls.option import OptionCoveralls


@pytest.fixture()
def opt():
    return OptionCoveralls('coveralls')


@pytest.fixture()
def cfg():
    return Config()


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 2


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 1



import logging
from pathlib import Path

from pkglts.option.reqs.option import requirements
from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionConda(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        LOGGER.info("update parameters %s", self._name)
        sec = dict(
            env_name="{{ base.pkgname }}",
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        env_name = cfg[self._name]['env_name']

        if not env_name:
            invalids.append("conda.env_name")

        if len(env_name) != len(env_name.strip()) or "-" in env_name:
            invalids.append("conda.env_name")

        return invalids

    def require_option(self, cfg):
        return ['pyproject']

    def environment_extensions(self, cfg):
        # use list(dict.fromkeys()) instead of set to preserve order
        # between two regenerate
        # requirements are already sorted by name
        channels = list(dict.fromkeys([dep.channel for dep in requirements(cfg) if dep.channel is not None]))
        channels_minimal = list(dict.fromkeys([dep.channel for dep in requirements(cfg)
                                               if 'install' in dep.intents and dep.channel is not None]))

        return {"channels": channels, "channels_minimal": channels_minimal}



import logging
from pathlib import Path

from url_normalize import url_normalize

from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionPyproject(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        LOGGER.info("update parameters %s", self._name)
        sec = dict(
            intended_versions=["3.11"],
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        intended_versions = cfg[self._name]['intended_versions']

        if not intended_versions:
            invalids.append("pyproject.intended_versions")

        return invalids

    def require_option(self, cfg):
        return ['src', 'doc', 'license', 'version', 'reqs']

    def environment_extensions(self, cfg):
        py_vers = sorted([int(v) for v in ver.split(".")] for ver in cfg[self._name]['intended_versions'])

        universal = len(set(ver.split(".")[0] for ver in cfg[self._name]['intended_versions'])) > 1

        return {
            "urls": find_urls(cfg),
            "py_max_ver": ".".join(str(v) for v in py_vers[-1]),
            "py_min_ver": ".".join(str(v) for v in py_vers[0]),
            "universal": universal
        }


def find_urls(cfg):
    """Extract all valid urls from all config sections.

    Args:
        cfg (Config):  current package configuration

    Returns:
        dict(str, str): key, url
    """
    urls = {}
    for name, section, key in [
        ("homepage", "base", "url"),
        ("repository", "github", "url"),
        ("repository", "gitlab", "url"),
    ]:
        try:
            url = cfg[section][key]
            if url is not None:
                urls[name] = url_normalize(url)
        except KeyError:
            pass

    return urls



from pkglts.config_management import Config


def test_src_pth():
    cfg = Config({
        'base': {'pkgname': 'toto', 'namespace': None},
        'src': {'namespace_method': 'pkg_util'}
    })
    cfg.load_extra()
    assert cfg._env.globals['src'].src_pth == "src/toto"



import logging
from pathlib import Path

from pkglts.option.doc.badge import Badge
from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionLgtm(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['github', 'pyproject']

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        url = f"https://lgtm.com/projects/g/{owner}/{project}/context:python"
        img = f"https://img.shields.io/lgtm/grade/python/g/{owner}/{project}.svg?logo=lgtm&logoWidth=18"
        badge = Badge(
            name="lgtm",
            url=url,
            url_img=img,
            text="Language grade: Python"
        )

        return {"badge": badge}



"""Regroup set of functions that make use of local environment
inside a package. Just a way to normalize pre defined paths.
"""
from pathlib import Path

from .hash_management import pth_as_key


def pkg_full_name(cfg):
    """Compute name of src dir according to pkgname
    and namespace in environment

    Args:
        cfg (Config):  current package configuration

    Returns:
        (str)
    """
    namespace = cfg['base']['namespace']
    if namespace is None:
        return cfg['base']['pkgname']

    return namespace + "." + cfg['base']['pkgname']


def src_dir(cfg):
    """Compute name of src dir according to pkgname
    and namespace in environment

    Args:
        cfg (Config):  current package configuration

    Returns:
        (Path)
    """
    rep = Path("src")
    namespace = cfg['base']['namespace']
    if namespace is not None:
        rep /= namespace

    return rep / cfg['base']['pkgname']


def init_namespace_dir(pth, rg_tree):
    """Populate a directory with specific __init__.py
    for namespace packages.

    Args:
        pth (Path): path in which to create the files
        rg_tree (dict): Structure to store path to templates found

    Returns:
        None
    """
    src_pth = Path(__file__).parent / "resource/namespace__init__.py.tpl"
    tgt_pth = pth / "__init__.py"
    rg_tree[pth_as_key(tgt_pth)] = [src_pth]



from pkglts.dependency import Dependency


def test_dependency_respect_strict_pkg_mng():
    # conda
    dep = Dependency("toto", pkg_mng="conda")
    assert dep.is_conda(strict=True)
    assert dep.is_conda(strict=False)
    dep = Dependency("toto", pkg_mng=None)
    assert not dep.is_conda(strict=True)
    assert dep.is_conda(strict=False)
    dep = Dependency("toto", pkg_mng="pip")
    assert not dep.is_conda(strict=True)
    assert not dep.is_conda(strict=False)
    dep = Dependency("toto", pkg_mng="git+https://")
    assert not dep.is_conda(strict=True)
    assert not dep.is_conda(strict=False)

    # pip
    dep = Dependency("toto", pkg_mng="pip")
    assert dep.is_pip(strict=True)
    assert dep.is_pip(strict=False)
    dep = Dependency("toto", pkg_mng=None)
    assert not dep.is_pip(strict=True)
    assert dep.is_pip(strict=False)
    dep = Dependency("toto", pkg_mng="conda")
    assert not dep.is_pip(strict=True)
    assert not dep.is_pip(strict=False)
    dep = Dependency("toto", pkg_mng="git+https://")
    assert not dep.is_pip(strict=True)
    assert not dep.is_pip(strict=False)


def test_dependency_writes_correct_pip_requirements():
    # no version specified
    dep = Dependency("toto", pkg_mng="pip")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto  # pip install toto"

    # normalization of package names
    dep = Dependency("toto_tutu", pkg_mng="pip")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto-tutu"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto-tutu  # pip install toto-tutu"

    # version fully specified without comparison indicator
    dep = Dependency("toto", pkg_mng="pip", version="2.18.0")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto==2.18.0"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto==2.18.0  # pip install toto==2.18.0"

    # version partially specified without comparison indicator
    dep = Dependency("toto", pkg_mng="pip", version="2.18")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto>=2.18, <2.19"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto>=2.18, <2.19  # pip install toto>=2.18, <2.19"

    # version specified with comparison indicator
    dep = Dependency("toto", pkg_mng="pip", version="==2.18")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto==2.18"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto==2.18  # pip install toto==2.18"

    # no version specified conda manager
    dep = Dependency("toto", pkg_mng="conda")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto  # conda install toto"

    # version fully specified without comparison indicator conda manager
    dep = Dependency("toto", pkg_mng="conda", version="2.18.1")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto==2.18.1"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto==2.18.1  # conda install toto=2.18.1"

    # version partially specified without comparison indicator conda manager
    dep = Dependency("toto", pkg_mng="conda", version="2.18")
    txt = dep.fmt_pip_requirement()
    assert txt == "toto>=2.18, <2.19"
    txt = dep.fmt_pip_requirement(extended=True)
    assert txt == "toto>=2.18, <2.19  # conda install toto=2.18"


def test_dependency_writes_correct_conda_requirements():
    # no version specified
    dep = Dependency("toto", pkg_mng="conda")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto"

    # normalization of package names
    dep = Dependency("toto_tutu", pkg_mng="conda")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto-tutu"

    # version partially specified no comparison operator
    dep = Dependency("toto", pkg_mng="conda", version="2.18")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto=2.18"

    # version fully specified no comparison operator
    dep = Dependency("toto", pkg_mng="conda", version="2.18.1")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto=2.18.1"

    # version specified conda comparison operator
    dep = Dependency("toto", pkg_mng="conda", version="=2.18")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto=2.18"

    # version specified pip comparison operator
    dep = Dependency("toto", pkg_mng="conda", version="==2.18")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto=2.18"

    # no version specified pip pkg_mng
    dep = Dependency("toto", pkg_mng="pip")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto"

    # version partially specified no comparison operator pip pkg_mng
    dep = Dependency("toto", pkg_mng="pip", version="2.18")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto=2.18"

    # version partially specified pip comparison operator pip pkg_mng
    dep = Dependency("toto", pkg_mng="pip", version="==2.18")
    txt = dep.fmt_conda_requirement()
    assert txt == "toto=2.18"



import logging
from pathlib import Path

from pkglts.option.doc.badge import Badge
from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionRequires(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['doc', 'github']

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        base_url = f"requires.io/github/{owner}/{project}/"
        url = f"{base_url}requirements/?branch=master"
        img = f"{base_url}requirements.svg?branch=master"
        badge = Badge(
            name="requires",
            url=url,
            url_img=img,
            text="Requirements status"
        )

        return {"badge": badge}



"""
This tool will display all dependencies used by this package.
"""
import ast
import logging
from pathlib import Path

from pkglts.local import src_dir

LOGGER = logging.getLogger(__name__)

stdpkgs = set(line.strip() for line in open(Path(__file__).parent / "stdpkgs.txt").readlines()
              if len(line.strip()) > 0 and not line.startswith("#"))


def iter_ext_imports(body):
    for node in body:
        if isinstance(node, ast.Import):
            for alias in node.names:
                yield alias.name
        elif isinstance(node, ast.ImportFrom):
            if node.level == 0:  # only external imports
                yield node.module

        # elif isinstance(node, ast.FunctionDef):
        #     for name, level in iter_imports(node.body):
        #         yield name, level


def find_reqs(pth):
    """Find all requirements (imports) used by a script.

    Args:
        pth (Path): path to file to parse

    Returns:
        (list): list of package names
    """
    pt = ast.parse(pth.read_bytes(), str(pth))
    # TODO pb with namespaces
    pkgs = set(pkgname.split(".")[0] for pkgname in iter_ext_imports(pt.body))

    return pkgs


def _iter_find_reqs(fld):
    for pth in fld.glob("*.py"):
        if not pth.name.startswith("_"):
            for req in find_reqs(pth):
                yield req

    for subfld in fld.glob("*/"):
        if subfld.name[0] not in ["_", "."]:
            for req in _iter_find_reqs(subfld):
                yield req


def action_find_reqs(cfg, **kwds):
    """Find dependencies used by this package.
    """
    LOGGER.info("Find requirements")
    this_pkgname = cfg['base']['pkgname']

    for dname in (src_dir(cfg), 'doc', 'example', 'script', 'test'):
        dirpth = Path(dname)
        if dirpth.exists():
            print(f"############\n# {dirpth}\n############")
            reqs = set(_iter_find_reqs(dirpth))
            reqs -= {this_pkgname}
            print("standard", sorted(reqs & stdpkgs))
            print("external", sorted(reqs - stdpkgs))
            print("")


def parser_find_reqs(subparsers):
    """Associate a CLI to this tool.

    Notes: The CLI will be a subcommand of pmg.

    Args:
        subparsers (ArgumentParser): entity to create a subparsers

    Returns:
        (string): a unique id for this parser
        (callable): the action to perform
    """
    subparsers.add_parser('reqs', help=action_find_reqs.__doc__)

    return 'reqs', action_find_reqs



from copy import deepcopy
from pathlib import Path

import pytest
from pkglts import logging_tools
from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.manage import install_example_files
from pkglts.small_tools import ensure_created, rmdir


@pytest.fixture()
def tmp_dir():
    pth = Path("tmp_cfgex")
    ensure_created(pth)

    yield pth

    rmdir(pth)


def test_install_example_returns_false_if_option_not_already_installed(tmp_dir):
    cfg = Config(DEFAULT_CFG)
    cfg.load_extra()
    ans = install_example_files('option', cfg, tmp_dir)
    assert not ans


def test_install_example_ok_if_option_do_not_provide_examples(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    ans = install_example_files('base', cfg, tmp_dir)
    assert not ans


def test_install_example_copy_files(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    pkg_cfg['test'] = dict(suite_name='nose')
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    assert len(tuple(tmp_dir.iterdir())) == 0
    logging_tools.main(2)
    install_example_files('test', cfg, tmp_dir)
    assert len(tuple(tmp_dir.iterdir())) > 0
    assert (tmp_dir / "src/toto/example.py").exists()
    assert (tmp_dir / "test/test_example.py").exists()


def test_install_example_copy_binary_files(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    pkg_cfg['data'] = dict(filetype=[".png", ".ui"], use_ext_dir=True)
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    assert len(tuple(tmp_dir.iterdir())) == 0
    install_example_files('data', cfg, tmp_dir)
    assert len(tuple(tmp_dir.iterdir())) > 0
    assert (tmp_dir / "src/toto_data/ext_data.png").exists()


def test_install_example_do_not_complain_if_file_already_exists(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    pkg_cfg['test'] = dict(suite_name='nose')
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    assert install_example_files('test', cfg, tmp_dir)
    assert install_example_files('test', cfg, tmp_dir)


def test_install_example_handles_namespace(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace='oa')
    pkg_cfg['src'] = dict(namespace_method='pkg_utils')
    pkg_cfg['test'] = dict(suite_name='nose')
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    install_example_files('test', cfg, tmp_dir)
    assert (tmp_dir / "test/test_example.py").exists()
    assert not (tmp_dir / "src/toto/example.py").exists()
    assert (tmp_dir / "src/oa/toto/example.py").exists()



import errno
import os
import sys
from shutil import rmtree
from time import sleep


def ensure_created(dname):
    dname.mkdir(exist_ok=True)


def rmdir(dname):
    for i in range(5):
        if dname.exists():
            try:
                rmtree(dname)
                return
            except OSError:
                sleep(0.1)
        else:
            return

    raise OSError(f"unable to remove directory: '{dname}'")


def ensure_path(pth):
    pth.parent.mkdir(parents=True, exist_ok=True)


# code taken from
# https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta

# Sadly, Python fails to provide the following magic number for us.
ERROR_INVALID_NAME = 123
'''
Windows-specific error code indicating an invalid pathname.

See Also
----------
https://msdn.microsoft.com/en-us/library/windows/desktop/ms681382%28v=vs.85%29.aspx
    Official listing of all such codes.
'''


def is_pathname_valid(pathname):
    """Check pathname validity

    `True` if the passed pathname is a valid pathname for the current OS;
    `False` otherwise.

    Args:
        pathname (str): does not need to exists

    Returns:
        (bool)
    """
    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(pathname, str) or not pathname:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os.path.splitdrive(pathname)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = os.environ.get('HOMEDRIVE', 'C:') if sys.platform == 'win32' else os.path.sep
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in pathname.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                if hasattr(exc, 'winerror'):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError as exc:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.
    #
    # Did we mention this should be shipped with Python already?



from pkglts.config_management import Config


def test_contributors_with_failed_request(mocker):
    mocker.patch('subprocess.check_output', side_effect=KeyError)
    cfg = Config(dict(git={}))
    cfg.load_extra()
    assert "failed" in cfg._env.globals['git'].contributors[0][0]


def test_contributors_without_failed_request(mocker):
    commits = b'commit Author: a <toto@titi>\n, commit: Author: b <tata@titi>\n, commit: Author: b <tata@titi>\n'
    mocker.patch('subprocess.check_output', return_value=commits)
    cfg = Config(dict(git={}))
    cfg.load_extra()
    contributors = cfg._env.globals['git'].contributors
    assert len(contributors) == 2



import logging
from pathlib import Path

from pkglts.option.doc.badge import Badge
from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionLandscape(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['flake8', 'travis']

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        url = f"landscape.io/github/{owner}/{project}/master"
        img = f"{url}/landscape.svg?style=flat"
        badge = Badge(
            name="landscape",
            url=url,
            url_img=img,
            text="Code health status"
        )

        return {"badge": badge}



import pytest
from pkglts.config_management import Config
from pkglts.dependency import Dependency
from pkglts.option.reqs.option import OptionReqs, fmt_conda_reqs, fmt_pip_reqs, requirements


@pytest.fixture()
def opt():
    return OptionReqs('reqs')


@pytest.fixture()
def cfg():
    return Config()


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['reqs']) == 1


def test_config_check_pkg_names(opt, cfg):
    cfg['reqs'] = {'require': []}
    assert 'reqs.require' not in opt.check(cfg)

    cfg['reqs'] = {'require': [{'pkg_mng': 'walou', 'name': 'numpy'}]}
    assert 'reqs.require' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0


def test_require_puts_list_of_reqs_in_requirements(opt, cfg):
    opt.update_parameters(cfg)
    cfg['reqs']['require'].append({'pkg_mng': 'walou', 'name': 'numpy'})

    assert 'numpy' in (dep.name for dep in requirements(cfg))


def test_fmt_conda_reqs_works_correctly():
    assert fmt_conda_reqs([], ['install']) == ""

    d1 = Dependency('d1')
    dex = Dependency('dex', intent='example')
    dconda = Dependency('dconda')
    dpip = Dependency('dpip')

    cmd = fmt_conda_reqs([d1, dex, dconda, dpip], ['install'])

    assert cmd.strip() == "conda install d1 dconda dpip"

    d = Dependency('d', channel='extra')

    assert fmt_conda_reqs([d], ['install']) == "conda install -c extra d"

    d = Dependency('d', version='= 5')

    assert fmt_conda_reqs([d], ['install']) == 'conda install "d= 5"'


def test_fmt_pip_reqs_works_correctly():
    assert fmt_pip_reqs([], ['install']) == ""

    d1 = Dependency('d1')
    dex = Dependency('dex', intent='example', pkg_mng='pip')
    dconda = Dependency('dconda')
    dpip = Dependency('dpip', pkg_mng='pip')

    cmd = fmt_pip_reqs([d1, dex, dconda, dpip], ['install'])

    assert cmd.strip() == 'pip install "dpip"'

    d = Dependency('d', pkg_mng='pip', version='= 5')

    assert fmt_pip_reqs([d], ['install']) == 'pip install "d== 5"'



import logging
from pathlib import Path

from pkglts.option_object import Option
from pkglts.version import __version__
from pkglts.dependency import normalize
from pkglts.local import pkg_full_name

LOGGER = logging.getLogger(__name__)


class OptionCondaAlias(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['conda']

    def environment_extensions(self, cfg):
        return {"norm_name": normalize(pkg_full_name(cfg))}



import logging
from pathlib import Path

from pkglts.option_object import Option
from pkglts.version import __version__
from . import bump_version

LOGGER = logging.getLogger(__name__)


class OptionVersion(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        LOGGER.info("update parameters %s", self._name)
        sec = dict(
            major=0,
            minor=0,
            post=1,
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        major = cfg[self._name]['major']
        minor = cfg[self._name]['minor']
        post = cfg[self._name]['post']

        if not isinstance(major, int):
            invalids.append("version.major")
        if not isinstance(minor, int):
            invalids.append("version.minor")
        if not isinstance(post, int):
            invalids.append("version.post")

        return invalids

    def require_option(self, cfg):
        return ['src']

    def tools(self, cfg):
        del cfg
        yield bump_version.parser_bump



from dataclasses import dataclass

from url_normalize import url_normalize


@dataclass
class Badge:
    name: str
    url: str
    url_img: str
    text: str = ""

    def __post_init__(self):
        self.url = url_normalize(self.url)
        self.url_img = url_normalize(self.url_img)

    def format(self, doc_fmt):
        """Produce valid img hyperlink.

        Args:
            doc_fmt (str): doc format either 'rst' or 'md'

        Returns:
            (str)
        """
        if self.text == "":
            txt = self.name
        else:
            txt = self.text

        if doc_fmt == 'rst':
            return ("\n"
                    f".. image:: {self.url_img}\n"
                    f"    :alt: {txt}\n"
                    f"    :target: {self.url}\n"
                    )

        if doc_fmt == 'md':
            return f"[![{txt}]({self.url_img})]({self.url})"

        raise UserWarning(f"Unknown format '{doc_fmt}'")



from pkglts.config_management import Config


def test_badge():
    cfg = Config(dict(requires={},
                      doc={'fmt': 'rst'},
                      github={'owner': "moi", 'project': "project"}))
    cfg.load_extra()
    assert "requires" in cfg._env.globals['requires'].badge.name



import logging
from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)


class OptionPluginProject(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        LOGGER.info("update parameters %s", self._name)
        cfg[self._name] = {
            "plugin_name": "{{ base.pkgname }}"
        }

    def check(self, cfg):
        invalids = []
        plugin_name = cfg[self._name]['plugin_name']

        if "." in plugin_name:
            invalids.append('plugin_project.plugin_name')
        elif not plugin_name.isidentifier():
            invalids.append('plugin_project.plugin_name')

        return invalids

    def require_option(self, cfg):
        return ['pyproject', 'data', 'git']

    def require(self, cfg):
        del cfg
        yield Dependency('pkglts', intent='install', channel='revesansparole')
